<?php

namespace Drupal\commerce_heidelpay\Plugin\Commerce\PaymentGateway;

use heidelpayPHP\Heidelpay;
use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_payment\Entity\PaymentInterface;
use Drupal\commerce_payment\Exception\InvalidRequestException;
use Drupal\commerce_payment\PaymentMethodTypeManager;
use Drupal\commerce_payment\PaymentTypeManager;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\HasPaymentInstructionsInterface;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\OffsitePaymentGatewayBase;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Lock\LockBackendInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Routing\CurrentRouteMatch;
use heidelpayPHP\Constants\PaymentState;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;

/**
 * Provides the Heidelpay payment gateway.
 *
 * @CommercePaymentGateway(
 *   id = "commerce_heidelpay_offsite",
 *   label = "Heidelpay",
 *   display_label = "Pay with credit or debit card",
 *   forms = {
 *     "offsite-payment" = "Drupal\commerce_heidelpay\PluginForm\OffsiteRedirect\HeidelpayForm",
 *   },
 *   credit_card_types = {
 *     "sandbox",
 *     "live",
 *   },
 *   payment_method_types = {"credit_card"},
 *   credit_card_types = {
 *     "amex",
 *     "dinersclub",
 *     "discover",
 *     "jcb",
 *     "mastercard",
 *     "visa",
 *   },
 *   requires_billing_information = FALSE,
 * )
 */
class HeidelpayOffSite extends OffsitePaymentGatewayBase implements HasPaymentInstructionsInterface {

  /**
   * The current route match service.
   *
   * @var \Drupal\Core\Routing\CurrentRouteMatch
   */
  protected $currentRouteMatch;

  /**
   * The locking layer instance.
   *
   * @var \Drupal\Core\Lock\LockBackendInterface
   */
  protected $lock;

  /**
   * The logger.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /**
   * Constructs a new Heidelpay object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\commerce_payment\PaymentTypeManager $payment_type_manager
   *   The payment type manager.
   * @param \Drupal\commerce_payment\PaymentMethodTypeManager $payment_method_type_manager
   *   The payment method type manager.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   The time.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_channel_factory
   *   The logger channel factory.
   * @param \Drupal\Core\Lock\LockBackendInterface $lock
   *   The locking layer instance.
   * @param \Drupal\Core\Routing\CurrentRouteMatch $current_route
   *   The current route match service.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, EntityTypeManagerInterface $entity_type_manager, PaymentTypeManager $payment_type_manager, PaymentMethodTypeManager $payment_method_type_manager, TimeInterface $time, LoggerChannelFactoryInterface $logger_channel_factory, LockBackendInterface $lock, CurrentRouteMatch $current_route) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $entity_type_manager, $payment_type_manager, $payment_method_type_manager, $time);

    $this->currentRouteMatch = $current_route;
    $this->lock = $lock;
    $this->logger = $logger_channel_factory->get('commerce_heidelpay');
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('plugin.manager.commerce_payment_type'),
      $container->get('plugin.manager.commerce_payment_method_type'),
      $container->get('datetime.time'),
      $container->get('logger.factory'),
      $container->get('lock'),
      $container->get('current_route_match')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'private_key' => '',
      'public_key' => '',
      'currency' => 'EUR',
      'instructions' => [
        'value' => '',
        'format' => 'plain_text',
      ],
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildConfigurationForm($form, $form_state);

    $form['private_key'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Private Key'),
      '#default_value' => $this->configuration['private_key'],
      '#size' => 60,
    ];
    $form['public_key'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Public Key'),
      '#default_value' => $this->configuration['public_key'],
      '#size' => 60,
    ];
    $form['currency'] = [
      '#type' => 'commerce_entity_select',
      '#title' => $this->t('currency'),
      '#default_value' => $this->configuration['currency'],
      '#target_type' => 'commerce_currency',
      '#hide_single_entity' => FALSE,
      '#multiple' => FALSE,
      '#required' => TRUE,
    ];
    $form['instructions'] = [
      '#type' => 'text_format',
      '#title' => $this->t('Payment instructions'),
      '#description' => $this->t('Shown the end of checkout, after the customer has placed their order.'),
      '#default_value' => $this->configuration['instructions']['value'],
      '#format' => $this->configuration['instructions']['format'],
    ];

    // @TODO Make the exclude payment methods config.
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::submitConfigurationForm($form, $form_state);

    if (!$form_state->getErrors()) {
      $values = $form_state->getValue($form['#parents']);

      $this->configuration['private_key'] = $values['private_key'];
      $this->configuration['public_key'] = $values['public_key'];
      $this->configuration['currency'] = $values['currency'];
      $this->configuration['instructions'] = $values['instructions'];
    }
  }

  /**
   * {@inheritdoc}
   */
  public function buildPaymentInstructions(PaymentInterface $payment) {
    $configuration = $this->getConfiguration();

    $instructions = [];
    if (!empty($configuration['instructions']['value'])) {
      $instructions = [
        '#type' => 'processed_text',
        '#text' => $configuration['instructions']['value'],
        '#format' => $configuration['instructions']['format'],
      ];
    }

    return $instructions;
  }

  /**
   * {@inheritdoc}
   */
  public function onReturn(OrderInterface $order, Request $request) {
    $data = $order->getData('commerce_heidelpay');
    if (empty($data['payment_id'])) {
      throw new InvalidRequestException($this->t('An error has been occurred trying of process the payment data, please contact with us.'));
    }

    $this->processNotification($order, $data['payment_id']);
  }

  /**
   * {@inheritdoc}
   */
  public function onNotify(Request $request) {
    // @TODO.
  }

  /**
   * Processes the notification.
   *
   * @param \Drupal\commerce_order\Entity\OrderInterface $order
   *   The order.
   * @param string $payment_id
   *   The external payment id.
   */
  protected function processNotification(OrderInterface $order, $payment_id) {
    try {
      $heidelpay = new Heidelpay($this->getConfiguration()['private_key']);
      $heidelpay_payment = $heidelpay->fetchPaymentByOrderId($order->id());

      // Somehow 6 is not mapped on PaymentState::mapStateCodeToName and it's
      // the status that is returned when clicking "Back to Merchant".
      if ($heidelpay_payment->getState() == 6) {
        // @TODO should this be an error on the site?
        // @TODO Better handle these exceptions because the only error displayed is the one in the parent, maybe messages?
        throw new InvalidRequestException($this->t('There was an error on the payment form, please retry.'));
      }

      $heidelpay_payment_state = $heidelpay_payment->getStateName();
      if ($heidelpay_payment_state != PaymentState::STATE_NAME_COMPLETED) {
        throw new InvalidRequestException($this->t('An error has been occurred trying of process the payment data, please contact with us.'));
      }

      $payment = $this->entityTypeManager->getStorage('commerce_payment')->create([
        'state' => $heidelpay_payment_state,
        'amount' => $order->getTotalPrice(),
        'payment_gateway' => $this->parentEntity->id(),
        'order_id' => $order->id(),
        'sandbox' => $this->getMode() == 'sandbox',
        'remote_id' => $payment_id,
        'remote_state' => $heidelpay_payment_state,
        'authorized' => $this->time->getRequestTime(),
      ]);
      $payment->save();

    }
    catch (\Exception $e) {
      throw new InvalidRequestException($this->t('An error has been occurred trying of process the payment data, please contact with us.'));
    }
  }

}
