<?php

namespace Drupal\commerce_heidelpay\Plugin\Commerce\PaymentGateway;

use Drupal\commerce_payment\Exception\PaymentGatewayException;
use Drupal\Core\Session\AccountProxyInterface;
use heidelpayPHP\Exceptions\HeidelpayApiException;
use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_payment\CreditCard;
use Drupal\commerce_payment\Entity\PaymentInterface;
use Drupal\commerce_payment\Entity\PaymentMethodInterface;
use Drupal\commerce_payment\PaymentMethodTypeManager;
use Drupal\commerce_payment\PaymentTypeManager;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\OnsitePaymentGatewayBase;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\DrupalKernelInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Url;
use heidelpayPHP\Heidelpay;
use heidelpayPHP\Resources\Basket;
use heidelpayPHP\Resources\CustomerFactory;
use heidelpayPHP\Resources\EmbeddedResources\Address;
use heidelpayPHP\Resources\EmbeddedResources\BasketItem;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\HttpFoundation\RequestStack;

/**
 * Provides the Heidelpay payment gateway (On site).
 *
 * @CommercePaymentGateway(
 *   id = "commerce_heidelpay_onsite",
 *   label = "Heidelpay (On Site)",
 *   display_label = "Pay with credit or debit card",
 *   forms = {
 *     "add-payment-method" = "Drupal\commerce_heidelpay\PluginForm\Heidelpay\PaymentMethodAddForm",
 *   },
 *   payment_method_types = {"credit_card"},
 *   credit_card_types = {
 *     "amex",
 *     "mastercard",
 *     "visa",
 *   },
 *   js_library = "commerce_heidelpay/form",
 *   requires_billing_information = FALSE,
 *
 * )
 */
class HeidelpayOnSite extends OnsitePaymentGatewayBase implements HeidelpayOnSiteInterface {

  /**
   * The logger.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /**
   * The Drupal kernel.
   *
   * @var \Drupal\Core\DrupalKernelInterface
   */
  protected $kernel;

  /**
   * The request stack.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  protected $requestStack;

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountProxyInterface
   */
  protected $currentUser;

  /**
   * Constructs a new HeidelpayOnsite object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\commerce_payment\PaymentTypeManager $payment_type_manager
   *   The payment type manager.
   * @param \Drupal\commerce_payment\PaymentMethodTypeManager $payment_method_type_manager
   *   The payment method type manager.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   The time.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_channel_factory
   *   The logger channel factory.
   * @param \Drupal\Core\DrupalKernelInterface $drupalKernel
   *   The drupal kernel.
   * @param \Symfony\Component\HttpFoundation\RequestStack $request_stack
   *   The request stack.
   * @param \Drupal\Core\Session\AccountProxyInterface $current_user
   *   The current user account.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, EntityTypeManagerInterface $entity_type_manager, PaymentTypeManager $payment_type_manager, PaymentMethodTypeManager $payment_method_type_manager, TimeInterface $time, LoggerChannelFactoryInterface $logger_channel_factory, DrupalKernelInterface $drupalKernel, RequestStack $request_stack, AccountProxyInterface $current_user) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $entity_type_manager, $payment_type_manager, $payment_method_type_manager, $time);

    $this->logger = $logger_channel_factory;
    $this->kernel = $drupalKernel;
    $this->requestStack = $request_stack;
    $this->currentUser = $current_user;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('plugin.manager.commerce_payment_type'),
      $container->get('plugin.manager.commerce_payment_method_type'),
      $container->get('datetime.time'),
      $container->get('logger.factory'),
      $container->get('kernel'),
      $container->get('request_stack'),
      $container->get('current_user')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'private_key' => '',
      'public_key' => '',
      '3ds' => TRUE,
      'account_holder' => FALSE,
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildConfigurationForm($form, $form_state);

    $form['private_key'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Private Key'),
      '#default_value' => $this->configuration['private_key'],
      '#size' => 60,
    ];
    $form['public_key'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Public Key'),
      '#default_value' => $this->configuration['public_key'],
      '#size' => 60,
    ];
    $form['3ds'] = [
      '#type' => 'radios',
      '#title' => $this->t('3ds enabled'),
      '#default_value' => $this->configuration['3ds'],
      '#options' => [
        TRUE => $this->t('Enabled'),
        FALSE => $this->t('Disabled'),
      ],
      '#description' => $this->t('When enabled, the user will need to complete an additional 3ds validation.'),
    ];
    $form['account_holder'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Request account holder on checkout'),
      '#default_value' => $this->configuration['account_holder'],
      '#description' => $this->t('A field requesting the account holder will be displayed on checkout.'),
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::submitConfigurationForm($form, $form_state);

    if (!$form_state->getErrors()) {
      $values = $form_state->getValue($form['#parents']);

      $this->configuration['private_key'] = $values['private_key'];
      $this->configuration['public_key'] = $values['public_key'];
      $this->configuration['3ds'] = $values['3ds'];
      $this->configuration['account_holder'] = $values['account_holder'];
    }
  }

  /**
   * {@inheritdoc}
   */
  public function createPayment(PaymentInterface $payment, $capture = TRUE) {
    $this->assertPaymentState($payment, ['new']);
    $payment_method = $payment->getPaymentMethod();
    assert($payment_method instanceof PaymentMethodInterface);
    $this->assertPaymentMethod($payment_method);
    $order = $payment->getOrder();
    assert($order instanceof OrderInterface);
    try {
      $heidelpay = new Heidelpay($this->getConfiguration()['private_key']);
      $heidelpay_payment_type = $heidelpay->fetchPaymentType($payment_method->getRemoteId());
      $payment_data = $order->getData('commerce_heidelpay');
      if (empty($payment_data) || !isset($payment_data['payment_id']) || !isset($payment_data['transaction_id'])) {
        $checkout_url = Url::fromRoute('commerce_checkout.form', ['commerce_order' => $order->id()], ['absolute' => TRUE])->toString();

        // Add customer info from the billing profile, default to the user data.
        if ($billing_profile = $payment_method->getBillingProfile()) {
          $billing_address = $billing_profile->get('address')->first()->toArray();
          $customer = CustomerFactory::createCustomer($billing_address['given_name'], $billing_address['family_name']);
          $billing_address_heidelpay = (new Address())
            ->setStreet($billing_address['address_line1'] . ' , ' . $billing_address['address_line2'])
            ->setState($billing_address['administrative_area'])
            ->setZip($billing_address['postal_code'])
            ->setCity($billing_address['locality'])
            ->setCountry($billing_address['country_code']);
          $customer->setBillingAddress($billing_address_heidelpay);
        }
        else {
          $name = $this->currentUser->getDisplayName();
          $customer = CustomerFactory::createCustomer($name, $name);
        }
        $customer->setEmail($this->currentUser->getEmail());

        // Add basket/cart info to the order.
        $basket_items = [];
        foreach ($order->getItems() as $order_item) {
          // @TODO Add a more sophisticated breakdown of items, such as discounts, VAT, etc.
          $basket_item = new BasketItem($order_item->label(), $order_item->getTotalPrice()->getNumber(), $order_item->getUnitPrice()->getNumber(), $order_item->getQuantity());
          // This seems to be needed, otherwise the basket shows 0.
          $basket_item->setAmountGross($order_item->getTotalPrice()->getNumber());
          $basket_items[] = $basket_item;
        }
        $basket = new Basket($order->id(), $order->getTotalPrice()->getNumber(), $order->getTotalPrice()->getCurrencyCode(), $basket_items);

        $heidelpay_transaction = $heidelpay->charge($order->getTotalPrice()->getNumber(), $order->getTotalPrice()->getCurrencyCode(), $heidelpay_payment_type->getId(), $checkout_url, $customer, $order->id(), NULL, $basket, (bool) $this->getConfiguration()['3ds']);

        // Set the transaction and payment id in case of 3ds, so the process can
        // return to where it was and finish the transaction.
        $order->setData('commerce_heidelpay', [
          'payment_id' => $heidelpay_transaction->getPaymentId(),
          'transaction_id' => $heidelpay_transaction->getId(),
        ]);
        $order->save();
      }
      else {
        $transaction_payment = $heidelpay->fetchPayment($payment_data['payment_id']);
        $heidelpay_transaction = $transaction_payment->getCharge($payment_data['transaction_id']);
      }
      $should_redirect = !empty($heidelpay_transaction->getRedirectUrl());

      // Redirect to the 3ds page or to success depending on the state of the
      // transaction.
      switch (TRUE) {
        case (!$should_redirect && $heidelpay_transaction->isSuccess()):
          // Save the payment.
          $payment->setRemoteId($payment_data['payment_id']);
          $payment->save();
          break;

        case ((!$should_redirect && $heidelpay_transaction->isPending()) || $heidelpay_transaction->isError()):
          $this->logger->get('commerce_heidelpay')->warning($heidelpay_transaction->getMessage()->getCode());
          throw new PaymentGatewayException('There was an error processing the payment.');

        case ($should_redirect && $heidelpay_transaction->isPending()):
          // Redirect in case of 3ds.
          $response = new RedirectResponse($heidelpay_transaction->getRedirectUrl());
          // Save session and ensure that the kernel terminate subscribers can
          // run.
          $request = $this->requestStack->getCurrentRequest();
          $request->getSession()->save();
          $response->prepare($request);
          $this->kernel->terminate($request, $response);
          $response->send();
          exit();
      }
    }
    catch (HeidelpayApiException $e) {
      // @TODO: Review the different heidelpay errors & throw contextual errors.
      $this->logger->get('commerce_heidelpay')->error($e->getCode() . ': ' . $e->getMessage());
      throw new PaymentGatewayException('There was an error processing the payment.');
    }
  }

  /**
   * {@inheritdoc}
   */
  public function createPaymentMethod(PaymentMethodInterface $payment_method, array $payment_details) {
    $required_keys = [
      // The expected keys are payment gateway specific and usually match
      // the PaymentMethodAddForm form elements. They are expected to be valid.
      'heidelpay_resource_id',
    ];
    foreach ($required_keys as $required_key) {
      if (empty($payment_details[$required_key])) {
        throw new PaymentGatewayException(sprintf('$payment_details must contain the %s key.', $required_key));
      }
    }

    try {
      $heidelpay = new Heidelpay($this->getConfiguration()['private_key']);
      $payment_type = $heidelpay->fetchPaymentType($payment_details['heidelpay_resource_id']);

      $payment_method->card_type = $this->cardBrandEquivalences($payment_type->getBrand());
      $payment_method->card_number = $payment_type->getNumber();
      list($month, $year) = explode('/', $payment_type->getExpiryDate());
      $payment_method->card_exp_month = $month;
      $payment_method->card_exp_year = $year;
      $payment_method->setRemoteId($payment_details['heidelpay_resource_id']);
      $expires = CreditCard::calculateExpirationTimestamp($month, $year);
      $payment_method->setExpiresTime($expires);
      $payment_method->save();
    }
    catch (\Exception $e) {
      // @TODO: Review the different heidelpay errors & throw contextual errors.
      $this->logger->get('commerce_heidelpay')->error($e->getCode() . ': ' . $e->getMessage());
      throw new PaymentGatewayException('There was an error processing the payment.');
    }
  }

  /**
   * {@inheritdoc}
   */
  public function deletePaymentMethod(PaymentMethodInterface $payment_method) {
    $payment_method->delete();
  }

  /**
   * {@inheritdoc}
   */
  public function getPublicKey() {
    return $this->configuration['public_key'];
  }

  /**
   * Translates the Heidelpay card type to what Commerce understands.
   *
   * @param string $brand
   *   The brand of the card on Heidelpay.
   *
   * @return string
   *   The commerce card type.
   */
  protected function cardBrandEquivalences($brand) {
    $types = [
      'AMEX' => 'amex',
      'MASTER' => 'mastercard',
      'VISA' => 'visa',
      'VISAELECTRON' => 'visa',
      'MAESTRO' => 'mastercard',
    ];

    return $types[$brand];
  }

}
