<?php

namespace Drupal\commerce_heidelpay\PluginForm\OffsiteRedirect;

use Drupal\commerce\Response\NeedsRedirectException;
use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_payment\PluginForm\PaymentOffsiteForm as BasePaymentOffsiteForm;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Url;
use heidelpayPHP\Resources\Basket;
use heidelpayPHP\Resources\EmbeddedResources\BasketItem;
use heidelpayPHP\Resources\PaymentTypes\Paypage;
use Symfony\Component\DependencyInjection\ContainerInterface;

use heidelpayPHP\examples\ExampleDebugHandler;
use heidelpayPHP\Heidelpay;
use heidelpayPHP\Resources\CustomerFactory;

/**
 * Provides the Heidelpay class for the payment form.
 */
class HeidelpayForm extends BasePaymentOffsiteForm implements ContainerInjectionInterface {

  use StringTranslationTrait;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * The time service.
   *
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  protected $time;

  /**
   * Constructs a new HeidelpayForm object.
   *
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger service.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   The time service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   */
  public function __construct(MessengerInterface $messenger, TimeInterface $time, EntityTypeManagerInterface $entity_type_manager) {
    $this->entityTypeManager = $entity_type_manager;
    $this->messenger = $messenger;
    $this->time = $time;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('messenger'),
      $container->get('datetime.time'),
      $container->get('entity_type.manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $payment = $this->entity;
    $order = $payment->getOrder();
    $payment_gateway_plugin = $payment->getPaymentGateway()->getPlugin();
    $gateway_settings = $payment_gateway_plugin->getConfiguration();

    try {
      $heidelpay = new Heidelpay($gateway_settings['private_key']);
      // @TODO Enable/Disable debug mode.
      $heidelpay->setDebugMode(TRUE)->setDebugHandler(new ExampleDebugHandler());

      // @TODO: Customer should come from real data.
      $customer = CustomerFactory::createCustomer('NameTest', 'SurnameTest');
      $paypage = new Paypage(119.0, 'EUR', $form['#return_url']);

      $paypage->setLogoImage('https://dev.heidelpay.com/devHeidelpay_400_180.jpg')
        ->setFullPageImage('https://www.heidelpay.com/fileadmin/content/header-Imges-neu/Header_Phone_12.jpg')
        ->setShopName('This is my test ship')
        ->setShopDescription('Description of the shop')
        ->setTagline('Tagline of the shop')
        ->setOrderId($order->id())
        ->setTermsAndConditionUrl('https://www.heidelpay.com/en/')
        ->setPrivacyPolicyUrl('https://www.heidelpay.com/de/')
        ->setImprintUrl('https://www.heidelpay.com/it/')
        ->setHelpUrl('https://www.heidelpay.com/at/')
        ->setContactUrl('https://www.heidelpay.com/en/about-us/about-heidelpay/');

      // @TODO This should be configuration.
      $paypage->setExcludeTypes([
        'przelewy24',
        'ideal',
        'prepayment',
        'invoice',
        'sepa-direct-debit-guaranteed',
        'invoice-guaranteed',
        'giropay',
        'alipay',
        'eps',
        'pis',
        'wechatpay',
        'invoice-factoring',
      ]);

      // @TODO Create a real basket.
      $basketItem = new BasketItem('Donation', 100.0, 119.0, 1);
      // @TODO: This seems to be needed, otherwise the basket shows 0.
      $basketItem->setAmountGross(119.0);

      $basket = new Basket($order->id(), 119.0, 'EUR', [$basketItem]);

      $heidelpay->initPayPageCharge($paypage, $customer, $basket);

      $order->setData('commerce_heidelpay', [
        'payment_id' => $paypage->getPaymentId(),
      ]);
      $order->save();
    }
    catch (\Exception $exception) {
      watchdog_exception('commerce_heidelpay', $exception);
      $this->messenger->addError($this->t('An error has been occurred trying of process the payment data, please contact with us.'));
      return $this->redirectToPaymentInformationPane($order);
    }

    return $this->buildRedirectForm($form, $form_state, $paypage->getRedirectUrl(), [], self::REDIRECT_GET);
  }

  /**
   * Redirects to the payment information pane on error.
   *
   * @param \Drupal\commerce_order\Entity\OrderInterface $order
   *   The order.
   *
   * @see \Drupal\commerce_payment\Plugin\Commerce\CheckoutPane\PaymentProcess::getErrorStepId()
   *
   * @throws \Drupal\commerce\Response\NeedsRedirectException
   */
  protected function redirectToPaymentInformationPane(OrderInterface $order) {
    try {
      /** @var \Drupal\commerce_checkout\Plugin\Commerce\CheckoutFlow\CheckoutFlowWithPanesInterface $checkout_flow */
      $checkout_flow = $order->get('checkout_flow')->first()->get('entity')->getTarget()->getValue()->getPlugin();
      $step_id = $checkout_flow->getPane('payment_information')->getStepId();
      if ($step_id == '_disabled') {
        // Can't redirect to the _disabled step. This could mean that
        // isVisible() was overridden to allow PaymentProcess to be used without
        // a payment_information pane, but this method was not modified.
        throw new \RuntimeException('Cannot get the step ID for the payment_information pane. The pane is disabled.');
      }

      $checkout_flow->redirectToStep($step_id);
    }
    catch (\Exception $exception) {
      $redirect_url = Url::fromRoute('<front>', ['absolute' => TRUE])->toString();

      throw new NeedsRedirectException($redirect_url);
    }
  }

}
